# cython: language_level=3

"""Discord Rich Presence for Team Fortress 2"""

# TF2 Rich Presence
# https://github.com/Kataiser/tf2-rich-presence
#
# Copyright (C) 2019 Kataiser & https://github.com/Kataiser/tf2-rich-presence/contributors
# https://github.com/Kataiser/tf2-rich-presence/blob/master/LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import ctypes
import datetime
import gc
import os
import platform
import time
import traceback
from typing import Any, Dict, Set, Tuple, Union

import psutil
from colorama import Style
from discoIPC import ipc

import configs
import console_log
import gamemodes
import launcher
import localization
import logger
import processes
import server
import settings
import utils

__author__ = "Kataiser"
__copyright__ = "Copyright (C) 2019 Kataiser & https://github.com/Kataiser/tf2-rich-presence/contributors"
__license__ = "GPL-3.0"
__email__ = "Mecharon1.gm@gmail.com"


def launch():
    try:
        gc.disable()

        log_main: logger.Log = logger.Log()
        log_main.to_stderr = launcher.DEBUG
        log_main.info(f"Starting TF2 Rich Presence {launcher.VERSION}")

        app: TF2RichPresense = TF2RichPresense(log_main)
        app.run()
    except (KeyboardInterrupt, SystemExit):
        raise SystemExit
    except Exception:
        try:
            gc.enable()
            log_main.critical(traceback.format_exc())
        except NameError:
            pass  # the crash happened in logger.Log().__init__() and so log_main is unassigned

        raise


class TF2RichPresense:
    def __init__(self, log: Union[logger.Log, None] = None, set_process_priority: bool = True):
        if log:
            self.log: logger.Log = log
        else:
            self.log: logger.Log = logger.Log()
            self.log.error(f"Initialized main.TF2RichPresense without a log, defaulting to one at {self.log.filename}")

        # this is what gets modified and sent to Discord via discoIPC
        self.activity: Dict[str, Union[str, Dict[str, int], Dict[str, str]]] = \
            {'details': 'In menus',
             'timestamps': {'start': int(time.time())},
             'assets': {'small_image': 'tf2_icon_small', 'small_text': 'Team Fortress 2', 'large_image': 'main_menu', 'large_text': 'Main menu'},
             'state': ''}
        # TODO: make activity a flat class instead that can be converted to a dict in the above format
        self.old_activity1: Dict[str, Union[str, Dict[str, int], Dict[str, str]]] = {}  # for the console output
        self.old_activity2: Dict[str, Union[str, Dict[str, int], Dict[str, str]]] = {}  # for sending to Discord
        self.activity_translated: Dict[str, Union[str, Dict[str, int], Dict[str, str]]] = {}
        self.client_connected: bool = False
        self.rpc_client: Union[ipc.DiscordIPC, None] = None
        self.test_state: str = 'init'
        self.should_mention_discord: bool = True
        self.should_mention_tf2: bool = True
        self.should_mention_steam: bool = True
        self.last_notify_time: Union[float, None] = None
        self.has_checked_class_configs: bool = False
        self.process_scanner: processes.ProcessScanner = processes.ProcessScanner(self.log)
        self.loc: localization.Localizer = localization.Localizer(self.log, settings.get('language'))
        self.current_time_formatted: str = ""
        self.current_map: Union[str, None] = None  # don't trust this variable
        self.time_changed_map: float = time.time()
        self.has_seen_kataiser: bool = False
        self.console_log_mtime: Union[int, None] = None
        self.old_console_log_mtime: Union[int, None] = None
        self.old_console_log_interpretation: tuple = ('', '', '')
        self.loop_iteration: int = 0
        self.custom_functions = None
        self.valid_usernames: Set[str] = set()
        self.last_name_scan_time: float = time.time()  # close enough
        self.steam_config_mtimes: Dict[str, int] = {}
        self.cleanup_primed: bool = True
        self.slow_sleep_time: bool = False
        self.has_set_process_priority: bool = not set_process_priority
        self.kataiser_scan_loop: int = 0
        self.last_server_request_time: float = 0.0
        self.last_server_request_data: str = ''
        self.last_server_request_address: str = ''

        try:
            self.log.cleanup(20 if launcher.DEBUG else 10)
        except (FileNotFoundError, PermissionError):
            self.log.error(f"Couldn't clean up logs folder:\n{traceback.format_exc()}")

        self.log.debug(f"CPU: {psutil.cpu_count(logical=False)} cores, {psutil.cpu_count()} threads, {round(psutil.cpu_freq().max / 1000, 1)} GHz")

        platform_info: Dict[str, Any] = {'architecture': platform.architecture, 'machine': platform.machine, 'system': platform.system, 'platform': platform.platform,
                                         'processor': platform.processor, 'python_version_tuple': platform.python_version_tuple}
        for platform_part in platform_info:
            try:
                if platform_part == 'platform':
                    platform_info[platform_part] = platform_info[platform_part](aliased=True)
                else:
                    platform_info[platform_part] = platform_info[platform_part]()
            except Exception:
                self.log.error(f"Exception during platform.{platform_part}(), skipping\n{traceback.format_exc()}")
        self.log.debug(f"Platform: {platform_info}")

        if not os.path.supports_unicode_filenames:
            self.log.error("Looks like the OS doesn't support unicode filenames. This might cause problems")

        settings.fix_settings(self.log)  # this should've been already dealt with by init, but sometimes that seems to just not work
        default_settings: dict = settings.defaults()
        current_settings: dict = settings.access_registry()

        if current_settings == default_settings:
            self.log.debug("Current settings are default")
        else:
            self.log.debug(f"Non-default settings: {settings.compare_settings(default_settings, current_settings)}")

        self.import_custom()

    def __repr__(self) -> str:
        return f"main.TF2RichPresense (state={self.test_state})"

    # import custom functionality
    def import_custom(self):
        custom_functions_path: str = os.path.join('resources', 'custom.py') if os.path.isdir('resources') else 'custom.py'

        if os.path.isfile(custom_functions_path):
            with open(custom_functions_path, 'r') as custom_functions_file:
                custom_functions_lines: int = len(custom_functions_file.readlines())

            import custom
            self.log.debug(f"Imported custom.py ({custom_functions_lines} lines)")
            self.custom_functions = custom.TF2RPCustom()  # good naming
        else:
            self.log.debug("custom.py doesn't exist")

    def run(self):
        while True:
            self.loop_body()

            # rich presence only updates every 15 seconds, but it listens constantly so sending every 2 or 5 seconds (by default) is probably fine
            sleep_time: int = settings.get('wait_time_slow') if self.slow_sleep_time else settings.get('wait_time')
            self.log.debug(f"Sleeping for {sleep_time} seconds (slow = {self.slow_sleep_time})")
            time.sleep(sleep_time)

    # the main logic. runs every 2 seconds (by default)
    def loop_body(self):
        self.slow_sleep_time = False
        self.loop_iteration += 1
        self.log.debug(f"Main loop iteration this app session: {self.loop_iteration}")

        constant_dynamic_activity: bool = settings.get('second_line') != 'Class' and self.test_state == 'in game'  # test state can become out of date later this loop but it's right atm
        self.old_activity1 = copy.deepcopy(self.activity)
        self.old_activity2 = copy.deepcopy(self.activity)
        if constant_dynamic_activity and '?' not in self.old_activity1['state']:
            self.old_activity1['state'] = ''

        if self.custom_functions:
            self.custom_functions.before_loop(self)

        # this as a one-liner is beautiful :)
        p_data: Dict[str, Dict[str, Union[bool, str, int, None]]] = self.process_scanner.scan()

        # reads steam config files to find usernames with -condebug (on init, and if any of them have been modified)
        if p_data['Steam']['running']:
            config_scan_needed: bool = self.steam_config_mtimes == {}

            for steam_config in self.steam_config_mtimes:
                old_mtime: int = self.steam_config_mtimes[steam_config]
                new_mtime: int = int(os.stat(steam_config).st_mtime)

                if new_mtime > old_mtime:
                    self.log.debug(f"Rescanning Steam config files ({new_mtime} > {old_mtime} for {steam_config})")
                    config_scan_needed = True

            if config_scan_needed:
                self.valid_usernames.update(self.steam_config_file(p_data['Steam']['path'], p_data['TF2']['running']))
                self.log.debug(f"Usernames with -condebug: {self.valid_usernames}")
        elif p_data['Steam']['pid'] is not None or p_data['Steam']['path'] is not None:
            self.log.error(f"Steam isn't running but its process info is {p_data['Steam']}. WTF?")

        # used for display only
        current_time: str = datetime.datetime.now().strftime('%I:%M:%S %p')
        self.current_time_formatted = current_time[1:] if current_time.startswith('0') else current_time

        if p_data['TF2']['running'] and p_data['Discord']['running'] and p_data['Steam']['running']:
            if not p_data['Steam']['running'] and p_data['TF2']['running']:
                self.log.error("TF2 is running but Steam isn't. WTF?")

            if not self.has_checked_class_configs:
                # modifies a few tf2 config files
                configs.class_config_files(self.log, p_data['TF2']['path'])
                self.has_checked_class_configs = True

            console_log_path: str = os.path.join(p_data['TF2']['path'], 'tf', 'console.log')
            top_line: str
            bottom_line: str
            # TODO: use a state machine and/or much more consistent var names
            top_line, bottom_line, server_address = self.interpret_console_log(console_log_path, self.valid_usernames, tf2_start_time=p_data['TF2']['time'])
            self.old_console_log_mtime = self.console_log_mtime
            self.old_console_log_interpretation = (top_line, bottom_line, server_address)

            actual_current_class: str = bottom_line  # https://www.portal2sounds.com/1130
            queued_in_game: bool = bottom_line not in console_log.tf2_classes and bottom_line != 'unselected'  # ditto

            if 'In menus' in top_line:
                # in menus displays the main menu
                self.test_state = 'menus'
                self.current_map = None
                self.activity['assets']['small_image'] = 'tf2_icon_small'
                self.activity['assets']['small_text'] = 'Team Fortress 2'

                if bottom_line == 'Queued for Casual':
                    self.activity['assets']['large_image'] = 'casual'
                    self.activity['assets']['large_text'] = bottom_line
                elif bottom_line == 'Queued for Competitive':
                    self.activity['assets']['large_image'] = 'comp'
                    self.activity['assets']['large_text'] = bottom_line
                elif 'Queued for MvM' in bottom_line:
                    self.activity['assets']['large_image'] = 'mvm_queued'
                    self.activity['assets']['large_text'] = bottom_line
                else:
                    self.activity['assets']['large_image'] = 'main_menu'
                    self.activity['assets']['large_text'] = 'Main menu'
            else:  # not in menus = in a game
                self.test_state = 'in game'
                class_pic_type: str = settings.get('class_pic_type').lower()
                second_line_setting: str = settings.get('second_line')
                class_line: str = self.loc.text("Class: {0}").format(self.loc.text(bottom_line)) if not queued_in_game else self.loc.text(bottom_line)

                if class_pic_type == 'none, use tf2 logo' or bottom_line == 'unselected' or queued_in_game:
                    self.activity['assets']['small_image'] = 'tf2_icon_small'
                    self.activity['assets']['small_text'] = 'Team Fortress 2'
                else:
                    # btw it's not possible to set the correct class icon while in a game and queued with the current setup
                    small_class_image = f'{bottom_line.lower()}_{class_pic_type}'
                    self.log.debug(f"Setting class small image to {small_class_image}")

                    self.activity['assets']['small_image'] = small_class_image
                    self.activity['assets']['small_text'] = bottom_line

                if second_line_setting in ('Player count', 'Kills'):
                    bottom_line = self.get_match_data(server_address, second_line_setting)
                elif second_line_setting == 'Class':
                    bottom_line = class_line
                elif second_line_setting == 'Time on map':
                    if self.current_map != top_line:  # top_line means the current map here... I should probably refactor that
                        self.current_map = top_line
                        self.time_changed_map = time.time()

                    # convert seconds to a pretty timestamp
                    seconds_on_map: float = time.time() - self.time_changed_map
                    time_format: str = '%M:%S' if seconds_on_map <= 3600 else '%H:%M:%S'
                    map_time_formatted: str = time.strftime(time_format, time.gmtime(seconds_on_map))

                    # I know I could just set the start time in activity, but I'd rather that always meant time with the game open
                    bottom_line = self.loc.text("Time on map: {0}").format(map_time_formatted)
                else:
                    self.log.error(f"Second line is invalid ({second_line_setting})")

                # good code
                hosting: bool = ' (hosting)' in top_line
                top_line = top_line.replace(' (hosting)', '')

                map_fancy, current_gamemode, gamemode_fancy = gamemodes.get_map_gamemode(self.log, top_line)
                map_out: str = map_fancy
                self.activity['assets']['large_image'] = current_gamemode
                self.activity['assets']['large_text'] = gamemode_fancy

                top_line = self.loc.text("Map: {0}").format(map_out)
                top_line = f"{top_line}{self.loc.text(' (hosting)')}" if hosting else top_line

            if not top_line:
                self.log.error("Top line is blank")
            if not bottom_line:
                self.log.error("Bottom line is blank")

            self.activity['details'] = top_line
            self.activity['state'] = bottom_line
            og_large_text: str = self.activity['assets']['large_text']  # why
            self.activity['assets']['large_text'] = self.loc.text(self.activity['assets']['large_text'])
            self.activity['timestamps']['start'] = p_data['TF2']['time']

            if self.custom_functions:
                self.custom_functions.loop_middle(self)

            activity_comparison: Dict[str, Union[str, Dict[str, int], Dict[str, str]]] = copy.deepcopy(self.activity)
            if constant_dynamic_activity and '?' not in activity_comparison['state']:
                activity_comparison['state'] = ''

            if activity_comparison != self.old_activity1:
                # output to terminal, just for monitoring
                print(f"{self.current_time_formatted}{utils.generate_delta(self.loc, self.last_notify_time)}{Style.BRIGHT}")

                base_window_title: str = self.loc.text("TF2 Rich Presence ({0})").format(launcher.VERSION)
                window_title_format_menus: str = self.loc.text("{0} - {1} ({2})")
                window_title_format_main: str = self.loc.text("{0} - {1} on {2}")

                if [d for d in ('Queued', 'Main menu') if d in og_large_text]:
                    # if queued or on the main menu, simplify cmd output
                    print(self.loc.text(self.activity['details']))
                    print(self.loc.text(self.activity['state']))
                    window_title = window_title_format_menus.format(base_window_title, self.loc.text(self.activity['details']), self.loc.text(self.activity['state']))
                else:
                    print(f"{self.loc.text(self.activity['details'])} ({self.activity['assets']['large_text']})")
                    print(self.loc.text(self.activity['state']))

                    if settings.get('second_line') != 'Class':
                        print(class_line)  # this means the current class. god this desperately needs a refactor

                    window_title = window_title_format_main.format(base_window_title, actual_current_class, map_out)

                print(Style.RESET_ALL, end='')

                time_elapsed: str = str(datetime.timedelta(seconds=int(time.time() - p_data['TF2']['time'])))
                print(self.loc.text("{0} elapsed").format(time_elapsed[2:] if time_elapsed.startswith('0:') else time_elapsed))
                print()

                self.log.info(f"Activity changed, outputting (old: {self.old_activity1}, new: {self.activity})")
                self.last_notify_time = time.time()

                ctypes.windll.kernel32.SetConsoleTitleW(window_title)
                self.log.debug(f"Set window title to \"{window_title}\"")
            else:
                self.log.debug("Activity hasn't changed, not outputting")

            if self.activity != self.old_activity2:
                # send everything to discord
                large_text_base = self.activity['assets']['large_text']
                self.activity['assets']['large_text'] += self.loc.text(" - TF2 Rich Presence {0}").format(launcher.VERSION)
                self.send_rpc_activity()
                self.activity['assets']['large_text'] = large_text_base
                # this gets reset because the old activity doesn't have it
            else:
                self.log.debug("Activity hasn't changed, not sending to Discord")

            if not self.client_connected:
                self.log.critical("self.client is disconnected when it shouldn't be")

        elif not p_data['TF2']['running']:
            self.necessary_program_not_running('Team Fortress 2', self.should_mention_tf2, 'TF2')
            self.should_mention_tf2 = False
        elif not p_data['Discord']['running']:
            self.necessary_program_not_running('Discord', self.should_mention_discord)
            self.should_mention_discord = False
        else:
            # last but not least, Steam
            self.necessary_program_not_running('Steam', self.should_mention_steam)
            self.should_mention_steam = False

        if not gc.isenabled():
            gc.enable()
            gc.collect()
            self.log.debug("Enabled GC and collected")

        if not self.has_set_process_priority:
            self_process: psutil.Process = psutil.Process()
            priorities_before: tuple = (self_process.nice(), self_process.ionice())
            self_process.nice(psutil.BELOW_NORMAL_PRIORITY_CLASS)
            self_process.ionice(psutil.IOPRIO_LOW)
            priorities_after: tuple = (self_process.nice(), self_process.ionice())
            self.log.debug(f"Set process priorities from {priorities_before} to {priorities_after}")
            self.has_set_process_priority = True

        if self.custom_functions:
            self.custom_functions.after_loop(self)

        return self.client_connected, self.rpc_client

    # notify user (possibly) and restart (possibly)
    def necessary_program_not_running(self, program_name: str, should_mention: bool, name_short: str = ''):
        name_short = program_name if not name_short else name_short
        self.test_state = f'no {name_short.lower()}'
        self.slow_sleep_time = True  # update less often if not all programs are running

        if self.client_connected:
            try:
                self.log.debug("Disconnecting client")
                self.rpc_client.disconnect()  # doesn't work...
                client_state = (self.rpc_client.client_id, self.rpc_client.connected, self.rpc_client.ipc_path, self.rpc_client.pid, self.rpc_client.platform, self.rpc_client.socket)
                self.log.debug(f"client state after disconnect: {client_state}")
            except Exception as err:
                self.log.error(f"client error while disconnecting: {err}", reportable=False)

            self.log.info("Restarting")
            del self.log
            raise SystemExit  # ...but this does
        else:
            self.log.info(f"{name_short} isn't running (mentioning to user: {should_mention})")

            if should_mention:
                print(f'{self.current_time_formatted}{utils.generate_delta(self.loc, self.last_notify_time)}{Style.BRIGHT}')
                print(self.loc.text(f"{program_name} isn't running"))
                print(Style.RESET_ALL)

                self.last_notify_time = time.time()
                self.should_mention_discord = True
                self.should_mention_tf2 = True
                self.should_mention_steam = True

                base_window_title: str = self.loc.text("TF2 Rich Presence ({0})").format(launcher.VERSION)
                window_title: str = self.loc.text("{0} - Waiting for {1}").format(base_window_title, program_name)
                ctypes.windll.kernel32.SetConsoleTitleW(window_title)
                self.log.debug(f"Set window title to \"{window_title}\"")

        # to prevent connecting when already connected
        self.client_connected = False

    # reads a console.log and returns current map and class
    def interpret_console_log(self, *args, **kwargs) -> Tuple[str, str, str]:
        return console_log.interpret(self, *args, **kwargs)

    # reads steam's launch options save file to find usernames with -condebug
    def steam_config_file(self, *args, **kwargs) -> set:
        return configs.steam_config_file(self, *args, **kwargs)

    # get player count or user score (kills) from the game server
    def get_match_data(self, *args, **kwargs) -> str:
        return server.get_match_data(self, *args, **kwargs)

    # sends localized RPC data, connecting to Discord initially if need be
    def send_rpc_activity(self):
        try:
            if not self.client_connected:
                # connects to Discord and sends first status, starts on main menu
                self.rpc_client = ipc.DiscordIPC(utils.get_api_key('discord'))
                self.rpc_client.connect()
                client_state: Tuple[Any, bool, str, int, str, Any] = (
                    self.rpc_client.client_id, self.rpc_client.connected, self.rpc_client.ipc_path, self.rpc_client.pid, self.rpc_client.platform, self.rpc_client.socket)
                self.log.debug(f"Initial RPC client state: {client_state}")

            # localize activity
            self.activity_translated = copy.deepcopy(self.activity)
            self.activity_translated['details'] = self.loc.text(self.activity['details'])
            self.activity_translated['assets']['small_text'] = self.loc.text(self.activity['assets']['small_text'])
            self.activity_translated['assets']['large_text'] = self.loc.text(self.activity['assets']['large_text'])

            # stop DB.json spam as the map time increases
            if not (settings.get('second_line') != 'Class' and self.test_state == 'in game'):
                self.activity_translated['state'] = self.loc.text(self.activity['state'])

            self.rpc_client.update_activity(self.activity_translated)
            self.log.info(f"Sent over RPC: {self.activity_translated}")
            client_state = (self.rpc_client.client_id, self.rpc_client.connected, self.rpc_client.ipc_path, self.rpc_client.pid, self.rpc_client.platform, self.rpc_client.socket)
            self.log.debug(f"client state: {client_state}")
            self.client_connected = True
        except Exception as client_connect_error:
            if str(client_connect_error) in ("Can't send data to Discord via IPC.", "Can't connect to Discord Client."):
                # often happens when Discord is in the middle of starting up
                self.log.error(str(client_connect_error), reportable=False)

                print(f'{self.current_time_formatted}{Style.BRIGHT}')
                print(self.loc.text("Can't connect to Discord for Rich Presence."))
                print(Style.RESET_ALL)

                time.sleep(settings.get('wait_time'))
                del self.log
                raise SystemExit
            else:
                raise


if __name__ == '__main__':
    launch()
